-- Blowfish for Euphoria v2.0 by Alexander Toresson toressonodakra@swipnet.se 040128
include file.e
include binary.e
global constant num = 16,
                bfdat = "Blowfish.dat"
global constant ENC_OFF = 0, ENC_ON = 1, ENC_AUTO = 2
global constant BLOWFISH_EBC = 0, BLOWFISH_CBC = 1
sequence P, S
integer fn
integer encrypt, encmode
integer oldopen, oldclose, oldputs, oldprint, oldprintf, oldgets, oldgetc, oldseek, oldwhere
sequence buffers, filefns, filemodes, errors, poss
constant GET_EOF = -1
constant CHUNK = 100
constant M_ALLOC = 16

buffers = {}
filefns = {}
filemodes = {}
errors = {}
poss = {}

encrypt = 0
encmode = 1

P = repeat(0, num + 2)
S = repeat(repeat(0, 256), 4)

type positive_int(integer x)
    return x >= 1
end type

global function allocate(positive_int n)
-- Allocate n bytes of memory and return the address.
-- Free the memory using free() below.
    return machine_func(M_ALLOC, n)
end function

function get_bytes(integer fn, integer n)
-- Return a sequence of n bytes (maximum) from an open file.
-- If n > 0 and fewer than n bytes are returned,
-- you've reached the end of file.
-- This function is normally used with files opened in binary mode.
    sequence s
    integer c, first, last

    if n = 0 then
	return {}
    end if

    c = getc(fn)
    if c = GET_EOF then
	return {}
    end if

    s = repeat(c, n)

    last = 1
    while last < n do
	-- for speed, read a chunk without checking for EOF
	first = last+1
	last  = last+CHUNK
	if last > n then
	    last = n
	end if
	for i = first to last do
	    s[i] = getc(fn)
	end for
	-- check for EOF after each chunk
	if s[last] = GET_EOF then
	    -- trim the EOF's and return
	    while s[last] = GET_EOF do
		last -= 1
	    end while
	    return s[1..last]
	end if
    end while
    return s
end function

function int_to_bytes(atom x)
-- returns value of x as a sequence of 4 bytes
-- that you can poke into memory
--      {bits 0-7,  (least significant)
--       bits 8-15,
--       bits 16-23,
--       bits 24-31} (most significant)
-- This is the order of bytes in memory on 386+ machines.
    integer a,b,c,d

    a = remainder(x, #100)
    x = floor(x / #100)
    b = remainder(x, #100)
    x = floor(x / #100)
    c = remainder(x, #100)
    x = floor(x / #100)
    d = remainder(x, #100)
    return {a,b,c,d}
end function

atom mem
mem = allocate(4)

function bytes_to_int(sequence s)
-- converts 4-byte peek() sequence into an integer value
    if length(s) = 4 then
	poke(mem, s)
    else
	poke(mem, s[1..4]) -- avoid breaking old code
    end if
    return peek4u(mem)
end function

function sprint(object x)
-- Return the string representation of any Euphoria data object.
-- This is the same as the output from print(1, x) or '?', but it's
-- returned as a string sequence rather than printed.
    sequence s

    if atom(x) then
	return sprintf("%.10g", x)
    else
	s = "{"
	for i = 1 to length(x) do
	    s &= sprint(x[i])
	    if i < length(x) then
		s &= ','
	    end if
	end for
	s &= "}"
	return s
    end if
end function

function Wrap(atom temp)
    while temp >= 2147483648 do
        temp -= 4294967296
    end while
    while temp < -2147483648 do
        temp += 4294967296
    end while
--    temp = bits_to_int(int_to_bits(temp, 32))
    return temp
end function

function F(atom x)
integer a
integer b
integer c
integer d
atom e
    d = and_bits(x, #FF) + 1
    c = and_bits(x, #FF00) / #100 + 1
    b = and_bits(x, #FF0000) / #10000 + 1
    a = and_bits(x, #FF000000) / #1000000 + 1
    if d <= 0 then d += 256 end if
    if c <= 0 then c += 256 end if
    if b <= 0 then b += 256 end if
    if a <= 0 then a += 256 end if
    e = S[1][a] + S[2][b]
    e = Wrap(e)
    e = xor_bits(e, S[3][c])
    e += S[4][d]
    e = Wrap(e)
    return e
end function

function encipher_8byte(atom xl, atom xr)
atom temp
    for a = 1 to num do
        xl = xor_bits(xl, P[a])
        xr = xor_bits(F(xl), xr)
        temp = xl
        xl = xr
        xr = temp
    end for
    temp = xl
    xl = xr
    xr = temp
    xr = xor_bits(xr, P[num+1])
    xl = xor_bits(xl, P[num+2])
    if xl < 0 then
        xl += #100000000
    end if
    if xr < 0 then
        xr += #100000000
    end if
    return {xl, xr}
end function

function decipher_8byte(atom xl, atom xr)
atom temp
    for a = num + 2 to 3 by -1 do
        xl = xor_bits(xl, P[a])
        xr = xor_bits(F(xl), xr)
        temp = xl
        xl = xr
        xr = temp
    end for
    temp = xl
    xl = xr
    xr = temp
    xr = xor_bits(xr, P[2])
    xl = xor_bits(xl, P[1])
    if xl < 0 then
        xl += #100000000
    end if
    if xr < 0 then
        xr += #100000000
    end if
    return {xl, xr}
end function

global function Init_Blowfish(sequence key)
sequence bytes, ret
atom hd
atom ld
integer charnum
    fn = open(bfdat, "rb")
    if fn = -1 then return 0 end if
    for a = 1 = 1 to num + 2 do
        bytes = get_bytes(fn, 4)
        P[a] = bytes_to_int(bytes)
    end for
    for a = 1 to 4 do
        for b = 1 to 256 do
            bytes = get_bytes(fn, 4)
            S[a][b] = bytes_to_int(bytes)
        end for
    end for
    close(fn)
    charnum = 1
    for a = 1 to num + 2 do
        P[a] = 0
        for b = 1 to 4 do
            P[a] = or_bits(P[a] * 256, key[charnum])
            charnum += 1
            if charnum > length(key) then charnum = 1 end if
        end for
    end for
    hd = 0
    ld = 0
    for a = 1 to num + 2 by 2 do
        ret = encipher_8byte(ld, hd)
        P[a] = ret[1]
        P[a+1] = ret[2]
    end for
    for a = 1 to 4 do
        for b = 1 to 256 by 2 do
            ret = encipher_8byte(ld, hd)
            S[a][b] = ret[1]
            S[a][b+1] = ret[2]
        end for
    end for
    return 1
end function

global function Blowfish_Encipher_Bytes(sequence bytes, integer typ)
sequence output
integer padding
sequence sel, ret, oldsel
atom xl, xr
    padding = remainder(length(bytes), 8)
    if padding != 0 then
        padding = 8 - padding
    end if
    bytes &= repeat(0, padding)
    output = repeat(0, length(bytes))
    if typ = 1 then
        oldsel = repeat(0, 8)
    end if
    for a = 1 to length(bytes) by 8 do
        sel = bytes[a..a+7]
        if typ = 1 then
            sel = xor_bits(sel, oldsel)
        end if
        xl = bytes_to_int(sel[1..4])
        xr = bytes_to_int(sel[5..8])
        ret = encipher_8byte(xl, xr)
        output[a..a+3] = int_to_bytes(ret[1])
        output[a+4..a+7] = int_to_bytes(ret[2])
        if typ = 1 then
            oldsel = int_to_bytes(ret[1]) & int_to_bytes(ret[2])
        end if
    end for
    output = padding & output
    return output
end function

global function Blowfish_Decipher_Bytes(sequence bytes, integer typ)
sequence output
integer padding
sequence sel, ret, oldsel, sel2
atom xl, xr
    if length(bytes) = 0 then return {} end if
    padding = bytes[1]
    if length(bytes) = 1 then
        bytes = {}
    else
        bytes = bytes[2..length(bytes)]
    end if
    output = repeat(0, length(bytes))
    if typ = 1 then
        oldsel = repeat(0, 8)
    end if
    for a = 1 to length(bytes) by 8 do
        sel = bytes[a..a+7]
        xl = bytes_to_int(sel[1..4])
        xr = bytes_to_int(sel[5..8])
        ret = decipher_8byte(xl, xr)
        if typ = 1 then
            sel2 = int_to_bytes(ret[1]) & int_to_bytes(ret[2])
            sel2 = xor_bits(sel2, oldsel)
            output[a..a+7] = sel2
            oldsel = sel
        else
            output[a..a+3] = int_to_bytes(ret[1])
            output[a+4..a+7] = int_to_bytes(ret[2])
        end if
    end for
    if length(output) - padding < 1 then
        output = {}
    else
        output = output[1..length(output)-padding]
    end if
    return output
end function

global function Blowfish_Encipher_Sequence(sequence seq, integer typ)
    return Blowfish_Encipher_Bytes(EncodeObject(seq), typ)
end function

global function Blowfish_Decipher_Sequence(sequence seq, integer typ)
    return DecodeObject(Blowfish_Decipher_Bytes(seq, typ))
end function

global procedure Set_Encrypt(integer state)
    encrypt = state
end procedure

global procedure Set_Mode(integer mode)
    encmode = mode
end procedure

function read_block(integer fn)
object name, len2, data
atom len
    name = get_bytes(fn, 4)
    len2 = get_bytes(fn, 4)
    if atom(name) or length(name) < 4 or atom(len2) or length(len2) < 4 then
        return {"ERR ", ""}
    end if
    len = bytes_to_int(len2)
    data = get_bytes(fn, len)
    if atom(data) or length(data) < len then
        return {"ERR ", ""}
    end if
    return {name, data}
end function

type four_bytes(sequence x)
    return length(x) = 4
end type

procedure write_block(integer fn, four_bytes name, sequence data)
    puts(fn, name)
    puts(fn, int_to_bytes(length(data)))
    puts(fn, data)
end procedure

function read_file(integer fn)
sequence block
object data
    data = 0
    while 1 do
        block = read_block(fn)
        if compare(block[1], "ERR ") = 0 then
            return -1
        elsif compare(block[1], "ENDF") = 0 then
            return data
        elsif compare(block[1], "DATA") = 0 then
            data = block[2]
        end if
    end while
end function

function openid(sequence filename, sequence mode)
    return open(filename, mode)
end function

oldopen = routine_id("openid")

--function ReduceCRLF(sequence data)
--sequence out
--integer pos
--    pos = 1
--    while 1 do
--        if length(data) >= pos + 1 and data[pos] = 'r' and data[pos] = 'l' then
--        end if
--    end while
--end function

without warning

global function open(sequence filename, sequence mode)
object data
integer filefn, ok
    if encrypt = ENC_ON or encrypt = ENC_AUTO then
        filefn = call_func(oldopen, {filename, {mode[1]}})
        ok = 1
        if mode[1] = 'r' or mode[1] = 'u' or mode[1] = 'a' then
            data = read_file(filefn)
            if atom(data) then
                if encrypt = ENC_ON then
                    return -1
                else
                    ok = 0
                end if
            end if
        end if
        close(filefn)
        if ok then
            filefn = call_func(oldopen, {filename, {mode[1]}})
            filefns &= filefn
            buffers = append(buffers, "")
            filemodes = append(filemodes, {mode, encmode})
            errors = append(errors, "")
            poss = append(poss, 1)
            if mode[1] = 'r' or mode[1] = 'u' or mode[1] = 'a' then
                data = Blowfish_Decipher_Bytes(data, encmode)
                buffers[length(buffers)] = data
            end if
            if mode[1] = 'a' then
                poss[length(poss)] = length(buffers[length(buffers)]) + 1
            end if
            return filefn
        end if
    end if
    filefn = call_func(oldopen, {filename, mode})
    return filefn
end function

procedure closeid(integer fn)
    close(fn)
end procedure

oldclose = routine_id("closeid")

global procedure close(integer fn)
integer found
    found = find(fn, filefns)
    if found and filemodes[found][1][1] != 'r' then
        write_block(fn, "ATPE", "0.1")
        write_block(fn, "DATA", Blowfish_Encipher_Bytes(buffers[found], encmode))
        write_block(fn, "ENDF", "")
        filefns[found] = {}
    end if
    call_proc(oldclose, {fn})
end procedure

function insert(sequence x, sequence y, integer pos)
    if pos = length(x) + 1 then
        x &= y
    elsif pos > length(x) then
        x &= repeat(0, pos - length(x) - 1) & y
    elsif pos + length(y) - 1 >= length(x) and pos = 1 then
        x = y
    elsif pos + length(y) - 1 >= length(x) then
        x = x[1..pos-1]
        x &= y
    else
        x[pos..pos+length(y)-1] = y
    end if
    return x
end function

procedure putsid(integer fn, sequence x)
    puts(fn, x)
end procedure

oldputs = routine_id("putsid")

global procedure puts(integer fn, sequence x)
integer found
    found = find(fn, filefns)
    if found then
        buffers[found] = insert(buffers[found], x, poss[found])
        poss[found] += length(x)
    else
        call_proc(oldputs, {fn, x})
    end if
end procedure

procedure printid(integer fn, object x)
    print(fn, x)
end procedure

oldprint = routine_id("printid")

global procedure print(integer fn, object x)
integer found
    found = find(fn, filefns)
    if found then
        buffers[found] = insert(buffers[found], sprint(x), poss[found])
        poss[found] += length(x)
    else
        call_proc(oldprint, {fn, x})
    end if
end procedure

procedure printfid(integer fn, sequence st, object x)
    printf(fn, st, x)
end procedure

oldprintf = routine_id("printfid")

global procedure printf(integer fn, sequence st, object x)
integer found
    found = find(fn, filefns)
    if found then
        buffers[found] = insert(buffers[found], sprintf(st, x), poss[found])
        poss[found] += length(x)
    else
        call_proc(oldprint, {fn, st, x})
    end if
end procedure

function getsid(integer fn)
    return gets(fn)
end function

oldgets = routine_id("getsid")

global function gets(integer fn)
integer found
sequence line, st
    found = find(fn, filefns)
    if found then
        if poss[found] > length(buffers[found]) then
            return -1
        end if
        line = ""
        while 1 do
            if poss[found] > length(buffers[found]) then
                exit
            end if
            line &= buffers[found][poss[found]]
            if buffers[found][poss[found]] = '\n' then
                poss[found] += 1
                exit
            end if
            poss[found] += 1
        end while
        return line
    else
        st = call_func(oldgets, {fn})
        return st
    end if
end function

function getcid(integer fn)
    return getc(fn)
end function

oldgetc = routine_id("getcid")

global function getc(integer fn)
integer found, char
    found = find(fn, filefns)
    if found then
        if poss[found] > length(buffers[found]) then
            return -1
        end if
        char = buffers[found][poss[found]]
        poss[found] += 1
        return char
    else
        return call_func(oldgetc, {fn})
    end if
end function

function seekid(integer fn, integer pos)
    return seek(fn, pos)
end function

oldseek = routine_id("seekid")

global function seek(integer fn, integer pos)
integer found
    found = find(fn, filefns)
    if found then
        poss[found] = pos + 1
        if pos >= 0 then return 0 else return 1 end if
    else
        return call_func(oldseek, {fn, pos})
    end if
end function

function whereid(integer fn)
    return where(fn)
end function

oldwhere = routine_id("whereid")

global function where(integer fn)
integer found
    found = find(fn, filefns)
    if found then
        return poss[found] - 1
    else
        return call_func(oldwhere, {fn})
    end if
end function

with warning
